var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import { nanoid } from 'nanoid';
import { Queue } from 'schummar-queue';
import { calcNextRun } from './helpers';
const CANCELED = Symbol('canceled');
export class LocalJob {
    constructor(scheduler, implementation, _a = {}) {
        var { maxParallel = LocalJob.DEFAULT_MAX_PARALLEL, retryCount = scheduler.options.retryCount, retryDelay = scheduler.options.retryDelay, log = scheduler.options.log } = _a, otherOptions = __rest(_a, ["maxParallel", "retryCount", "retryDelay", "log"]);
        this.scheduler = scheduler;
        this.implementation = implementation;
        this.handles = new Set();
        this.hasShutDown = false;
        this.executionIds = new Map();
        this.options = Object.assign({ maxParallel, retryCount, retryDelay, log }, otherOptions);
        this.q = new Queue({ parallel: this.options.maxParallel });
        this.schedule();
    }
    schedule() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const { schedule } = this.options;
                if (!schedule)
                    return;
                const data = schedule.data;
                while (!this.hasShutDown) {
                    const nextRun = calcNextRun(schedule);
                    yield this.sleep(nextRun.getTime() - Date.now());
                    yield this.execute(...[data]);
                }
            }
            catch (e) {
                if (e !== CANCELED)
                    throw e;
            }
        });
    }
    execute(...[data, { delay = 0, executionId = nanoid() } = {}]) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const existing = this.executionIds.get(executionId);
                if (existing)
                    return existing;
                const promise = (() => __awaiter(this, void 0, void 0, function* () {
                    yield this.sleep(delay);
                    let attempt = 0, error;
                    while (!this.hasShutDown) {
                        try {
                            return yield this.q.schedule(() => this.implementation(data, { error, attempt }));
                        }
                        catch (e) {
                            error = e;
                            if (!this.hasShutDown && attempt < this.options.retryCount) {
                                attempt++;
                                yield this.sleep(this.options.retryDelay);
                            }
                            else {
                                throw error;
                            }
                        }
                    }
                    throw CANCELED;
                }))();
                this.executionIds.set(executionId, promise);
                return yield promise;
            }
            catch (e) {
                if (e !== CANCELED) {
                    this.options.log('error', 'Error in job execution:', e);
                }
                throw e;
            }
            finally {
                this.executionIds.delete(executionId);
            }
        });
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            this.hasShutDown = true;
            for (const handle of this.handles) {
                handle();
            }
        });
    }
    sleep(ms) {
        return new Promise((resolve, reject) => {
            const timeout = setTimeout(() => {
                this.handles.delete(handle);
                resolve();
            }, ms);
            const handle = () => {
                clearTimeout(timeout);
                reject(CANCELED);
            };
            this.handles.add(handle);
        });
    }
}
LocalJob.DEFAULT_MAX_PARALLEL = 1;
