var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
import { MongoClient } from 'mongodb';
import { DistributedJob } from './distributedJob';
import { sleep } from './helpers';
import { LocalJob } from './localJob';
const defaultLogger = (level, ...args) => {
    if (level === 'error' || level === 'warn') {
        console[level](...args);
    }
};
export class Scheduler {
    constructor(collection, _a = {}) {
        var { retryCount = Scheduler.DEFAULT_RETRY_COUNT, retryDelay = Scheduler.DEFAULT_RETRY_DELAY, lockDuration = Scheduler.DEFAULT_LOCK_DURATION, lockCheckInterval = Scheduler.DEFAULT_LOCK_CHECK_INTERVAL, log = defaultLogger } = _a, otherOptions = __rest(_a, ["retryCount", "retryDelay", "lockDuration", "lockCheckInterval", "log"]);
        this.distributedJobs = new Set();
        this.localJobs = new Set();
        this.hasShutDown = false;
        this.options = Object.assign({ retryCount, retryDelay, lockDuration, lockCheckInterval, log }, otherOptions);
        if (collection && 'uri' in collection) {
            this.collection = MongoClient.connect(collection.uri).then((client) => client.db(collection.db).collection(collection.collection));
        }
        else {
            this.collection = collection;
        }
    }
    watch() {
        var e_1, _a;
        return __awaiter(this, void 0, void 0, function* () {
            if (this.hasShutDown || this.stream) {
                this.options.log('info', `Stopping db watcher because ${this.hasShutDown ? 'shutdown' : 'already runs'}`);
                return;
            }
            try {
                const col = yield this.collection;
                if (!col) {
                    throw new Error('No db set up!');
                }
                this.options.log('debug', 'start db watcher');
                this.stream = col.watch([
                    {
                        $match: { operationType: { $in: ['insert', 'replace', 'update'] } },
                    },
                ], { fullDocument: 'updateLookup' });
                this.stream.once('resumeTokenChanged', () => {
                    this.options.log('debug', 'db watcher first token');
                    for (const job of this.distributedJobs)
                        job.changeStreamReconnected();
                });
                // When starting watching or after connection loss, force refresh
                const cursor = this.stream.stream();
                try {
                    for (var cursor_1 = __asyncValues(cursor), cursor_1_1; cursor_1_1 = yield cursor_1.next(), !cursor_1_1.done;) {
                        const change = cursor_1_1.value;
                        this.options.log('debug', 'db watcher change received', 'fullDocument' in change && change.fullDocument ? `${change.fullDocument.jobId} ${change.fullDocument.executionId}` : undefined);
                        if ('fullDocument' in change && change.fullDocument) {
                            for (const job of this.distributedJobs) {
                                if (job.jobId === change.fullDocument.jobId) {
                                    job.receiveUpdate(change.fullDocument);
                                }
                            }
                        }
                    }
                }
                catch (e_1_1) { e_1 = { error: e_1_1 }; }
                finally {
                    try {
                        if (cursor_1_1 && !cursor_1_1.done && (_a = cursor_1.return)) yield _a.call(cursor_1);
                    }
                    finally { if (e_1) throw e_1.error; }
                }
            }
            catch (e) {
                delete this.stream;
                if (this.hasShutDown)
                    return;
                this.options.log('warn', 'Change stream error:', e);
                yield sleep(10);
                this.watch();
            }
        });
    }
    addJob(jobId, implementation = null, options) {
        if (!this.collection)
            throw Error('No db set up!');
        const job = new DistributedJob(this, this.collection, jobId, implementation, options);
        this.distributedJobs.add(job);
        this.hasShutDown = false;
        this.watch();
        return job;
    }
    addLocalJob(implementation, options) {
        const job = new LocalJob(this, implementation, options);
        this.localJobs.add(job);
        return job;
    }
    clearDB() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.collection)
                throw Error('No db set up!');
            const col = yield this.collection;
            yield col.deleteMany({});
        });
    }
    clearJobs() {
        return __awaiter(this, void 0, void 0, function* () {
            yield Promise.all([...this.distributedJobs, ...this.localJobs].map((job) => job.shutdown()));
            this.distributedJobs.clear();
            this.localJobs.clear();
        });
    }
    shutdown() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            this.hasShutDown = true;
            (_a = this.stream) === null || _a === void 0 ? void 0 : _a.close();
            yield this.clearJobs();
        });
    }
}
Scheduler.DEFAULT_LOCK_DURATION = 5 * 60 * 1000; // 5 minutes
Scheduler.DEFAULT_LOCK_CHECK_INTERVAL = 60 * 1000; // 1 minute
Scheduler.DEFAULT_RETRY_COUNT = 10;
Scheduler.DEFAULT_RETRY_DELAY = 60 * 1000; // 1 minute
